// Constants
const int kNumLights = 4;
const float kFar = 500.0;

const vec3 ambientColor = vec3(1.0, 1.0, 1.0);
const vec3 specColor = vec3(1.0, 1.0, 1.0);

// G-buffer
uniform sampler2D texGPosition;
uniform sampler2D texGNormal;
uniform sampler2D texGAlbedo;
uniform sampler2D texGMaterial;

// Blurred SSAO texture
uniform sampler2D texSSAO;

// Shadow map
uniform samplerCube texShadowMap;

uniform bool ssaoEnabled;
uniform float ambientFactor;

uniform mat4 viewMatrix;

// Light colors and positions
uniform vec3 lightColors[kNumLights];
uniform vec3 lightPositions[kNumLights];
uniform float lightRadii[kNumLights];

// Fog
uniform vec3 fogColor;
uniform float fogDensity;

// Brightness (for fades)
uniform float brightness;

in vec2 texCoord;

out vec4 fragColor;

void main (void)
{
    // Retrieve position/normal/albedo color from textures
    vec3 position = texture(texGPosition, texCoord).rgb;
    vec3 worldPos = (inverse(viewMatrix) * vec4(position, 1.0)).xyz;
    vec3 normal = texture(texGNormal, texCoord).rgb;
    vec4 albedo = texture(texGAlbedo, texCoord);
    float ambientOcclusion = texture(texSSAO, texCoord).r;

    float specularFactor = texture(texGMaterial, texCoord).g;

    // -- Begin Fog -----------------------------------------------------------
    // Get the distance from view space position range
    float dist = length(position);
    float fogFactor = 1.0 / exp((dist * fogDensity) * (dist * fogDensity));
    fogFactor = clamp(fogFactor, 0.0, 1.0);
    // -- End Fog -------------------------------------------------------------

    // If lighting flag disabled, just use the albedo
    if (texture(texGMaterial, texCoord).r < 0.5)
    {
        fragColor = albedo * brightness;
        return;
    }

    // -- Begin Shadow Mapping ------------------------------------------------
    // HACK: light position 0 is always the shadow-mapped/omni-directional light source
    vec3 lightPos = lightPositions[0];

    // Direction vector into the cubemap
    vec3 cubemapLookup = lightPos - worldPos;

    // HACK (maybe???): Couldn't get shadow map to look right without this
    // I've probably screwed up the signed-ness of an axis somewhere
    cubemapLookup.x *= -1.0;

    // Shadow map contains shortest distance to the light for a given point
    float shortestDistanceToLight = texture(texShadowMap, cubemapLookup).r;

    // Actual distance to the light from this point in world space
    float distanceToLight = length(cubemapLookup);

    // De-normalize the depth buffer value
    shortestDistanceToLight *= kFar;

    // Offset to avoid surface acne
    float shadowOffset = 0.18;

    float shadowFactor = 1.0;

    // In shadow if we're further than the shortest distance to the light
    if (shortestDistanceToLight + shadowOffset < distanceToLight)
        shadowFactor = 0.0;
    // -- End Shadow Mapping --------------------------------------------------

    // -- Begin Phong lighting model -------------------------------------------
    vec3 ambient = albedo.rgb * ambientColor * ambientFactor;
    vec3 diffuse = vec3(0.0, 0.0, 0.0);
    vec3 specular = vec3(0.0, 0.0, 0.0);

    for (int i = 0; i < kNumLights; i++)
    {
        vec3 viewLightPosition = (viewMatrix * vec4(lightPositions[i], 1.0)).xyz;

        vec3 incident = normalize(viewLightPosition - position);
        float lambert = max(0.0, dot(incident, normal));
        lambert *= shadowFactor;

        float dist = length(viewLightPosition - position);
        float atten = 1.0 - clamp(dist / lightRadii[i], 0.0, 1.0);
        vec3 viewDir = normalize(-position);
        vec3 halfDir = normalize(incident + viewDir);

        float rFactor = max(0.0, dot(halfDir, normal));
        float sFactor = pow(rFactor, 32.0);

        diffuse += albedo.rgb * lightColors[i] * lambert * atten;

        vec3 specFromThisLight = lightColors[i] * sFactor * atten * specularFactor;

        // This light could be shadow mapped - apply shadow factor
        if (i == 0)
            specFromThisLight *= shadowFactor;

        specular += specFromThisLight;
    }
    // -- End Phong lighting model --------------------------------------------

    if (ssaoEnabled)
        ambient *= pow(ambientOcclusion, 2);

    fragColor = vec4(ambient + diffuse + specular, albedo.a);

    // Apply fog
    fragColor = mix(vec4(fogColor, 1.0), fragColor, fogFactor);

    // Apply brightness
    fragColor.rgb *= brightness;
}